@@ -2,7 +2,7 @@
 /*                                                                          */
 /*  The FreeType project -- a free and portable quality TrueType renderer.  */
 /*                                                                          */
-/*  Copyright 2005-2009, 2011-2013 by                                       */
+/*  Copyright (C) 2005-2019 by                                              */
 /*  D. Turner, R.Wilhelm, and W. Lemberg                                    */
 /*                                                                          */
 /*                                                                          */
@@ -11,8 +11,13 @@
 /****************************************************************************/
 
 
+#ifndef  _GNU_SOURCE
+#define  _GNU_SOURCE /* we use `strcasecmp' */
+#endif
+
 #include <ft2build.h>
 #include FT_FREETYPE_H
+#include FT_MODULE_H
 
 #include FT_CACHE_H
 #include FT_CACHE_MANAGER_H
@@ -20,6 +25,12 @@
 #include FT_BITMAP_H
 #include FT_FONT_FORMATS_H
 
+  /* error messages */
+#undef FTERRORS_H_
+#define FT_ERROR_START_LIST     {
+#define FT_ERRORDEF( e, v, s )  case v: str = s; break;
+#define FT_ERROR_END_LIST       default: str = "unknown error"; }
+
 #include "common.h"
 #include "ftcommon.h"
 
@@ -34,6 +45,9 @@
 #endif
 
 
+#define N_HINTING_ENGINES  2
+
+
   FT_Error  error;
 
 
@@ -60,7 +74,13 @@
   void
   PanicZ( const char*  message )
   {
-    fprintf( stderr, "%s\n  error = 0x%04x\n", message, error );
+    const FT_String  *str;
+
+
+    switch( error )
+    #include FT_ERRORS_H
+
+    fprintf( stderr, "%s\n  error = 0x%04x, %s\n", message, error, str );
     exit( 1 );
   }
 
@@ -75,23 +95,41 @@
 
 
   FTDemo_Display*
-  FTDemo_Display_New( grPixelMode  mode,
-                      int          width,
-                      int          height )
+  FTDemo_Display_New( const char*  dims )
   {
     FTDemo_Display*  display;
+    grPixelMode      mode;
     grSurface*       surface;
     grBitmap         bit;
+    int              width, height, depth = 24;
 
 
+    if ( sscanf( dims, "%dx%dx%d", &width, &height, &depth ) < 2 )
+      return NULL;
+
+    switch ( depth )
+    {
+    case 8:
+      mode = gr_pixel_mode_gray;
+      break;
+    case 15:
+      mode = gr_pixel_mode_rgb555;
+      break;
+    case 16:
+      mode = gr_pixel_mode_rgb565;
+      break;
+    case 32:
+      mode = gr_pixel_mode_rgb32;
+      break;
+    default:
+      mode = gr_pixel_mode_rgb24;
+      break;
+    }
+
     display = (FTDemo_Display *)malloc( sizeof ( FTDemo_Display ) );
     if ( !display )
       return NULL;
 
-    if ( mode != gr_pixel_mode_gray  &&
-         mode != gr_pixel_mode_rgb24 )
-      return NULL;
-
     grInitDevices();
 
     bit.mode  = mode;
@@ -117,6 +155,8 @@
     display->warn_color = grFindColor( display->bitmap,
                                        0xff, 0x00, 0x00, 0xff );
 
+    display->gamma = GAMMA;
+
     return display;
   }
 
@@ -140,32 +180,9 @@
   FTDemo_Display_Clear( FTDemo_Display*  display )
   {
     grBitmap*  bit   = display->bitmap;
-    int        pitch = bit->pitch;
 
 
-    if ( pitch < 0 )
-      pitch = -pitch;
-
-    if ( bit->mode == gr_pixel_mode_gray )
-      memset( bit->buffer,
-              display->back_color.value,
-              (unsigned int)( pitch * bit->rows ) );
-    else
-    {
-      unsigned char*  p = bit->buffer;
-      int             i;
-
-
-      for ( i = 0; i < pitch; i += 3, p += 3 )
-      {
-        p[0] = display->back_color.chroma[0];
-        p[1] = display->back_color.chroma[1];
-        p[2] = display->back_color.chroma[2];
-      }
-
-      for ( i = 1; i < bit->rows; i++, p += pitch )
-        memcpy( p, p - pitch, (size_t)pitch );
-    }
+    grFillRect( bit, 0, 0, bit->width, bit->rows, display->back_color );
   }
 
 
@@ -243,9 +260,11 @@
 
         if ( has_extension )
           memcpy( suffix, orig, 5 );
+        else
+          *suffix = '\0';
       }
 
-      if ( (*aface)->charmaps )
+      if ( (*aface)->charmaps && font->cmap_index < (*aface)->num_charmaps )
         (*aface)->charmap = (*aface)->charmaps[font->cmap_index];
     }
 
@@ -286,25 +305,24 @@
     if ( error )
       PanicZ( "could not initialize charmap cache" );
 
-    FT_Bitmap_New( &handle->bitmap );
+    FT_Bitmap_Init( &handle->bitmap );
 
     FT_Stroker_New( handle->library, &handle->stroker );
 
-    handle->encoding = FT_ENCODING_NONE;
+    handle->encoding = FT_ENCODING_ORDER;
 
-    handle->hinted    = 1;
-    handle->antialias = 1;
-    handle->use_sbits = 1;
-    handle->autohint  = 0;
-    handle->lcd_mode  = 0;
-    handle->color     = 1;
+    handle->hinted     = 1;
+    handle->use_sbits  = 1;
+    handle->use_color  = 1;
+    handle->use_layers = 1;
+    handle->autohint   = 0;
+    handle->lcd_mode   = LCD_MODE_AA;
 
     handle->use_sbits_cache = 1;
 
     /* string_init */
     memset( handle->string, 0, sizeof ( TGlyph ) * MAX_GLYPHS );
     handle->string_length = 0;
-    handle->string_reload = 1;
 
     return handle;
   }
@@ -352,7 +370,8 @@
   FT_Error
   FTDemo_Install_Font( FTDemo_Handle*  handle,
                        const char*     filepath,
-                       FT_Bool         outline_only )
+                       FT_Bool         outline_only,
+                       FT_Bool         no_instances )
   {
     static char   filename[1024 + 5];
     long          i, num_faces;
@@ -390,7 +409,7 @@
       error = FT_New_Face( handle->library, filename, -( i + 1 ), &face );
       if ( error )
         continue;
-      instance_count = face->style_flags >> 16;
+      instance_count = no_instances ? 0 : face->style_flags >> 16;
       FT_Done_Face( face );
 
       /* load face with and without named instances */
@@ -409,16 +428,6 @@
           continue;
         }
 
-        if ( handle->encoding != FT_ENCODING_NONE )
-        {
-          error = FT_Select_Charmap( face, handle->encoding );
-          if ( error )
-          {
-            FT_Done_Face( face );
-            continue;
-          }
-        }
-
         font = (PFont)malloc( sizeof ( *font ) );
 
         /* We allocate four more bytes since we want to attach an AFM */
@@ -428,8 +437,15 @@
         strcpy( (char*)font->filepathname, filename );
 
         font->face_index = ( j << 16 ) + i;
-        font->cmap_index = face->charmap ? FT_Get_Charmap_Index( face->charmap )
-                                         : 0;
+
+        if ( handle-> encoding != FT_ENCODING_ORDER                      &&
+             FT_Select_Charmap( face, (FT_Encoding)handle->encoding ) ==
+                                                               FT_Err_Ok )
+          font->cmap_index = FT_Get_Charmap_Index( face->charmap );
+        else
+          font->cmap_index = face->num_charmaps;  /* FT_ENCODING_ORDER */
+
+        font->palette_index = 0;
 
         if ( handle->preload )
         {
@@ -439,6 +455,7 @@
 
           if ( file == NULL )  /* shouldn't happen */
           {
+            free( (void*)font->filepathname );
             free( font );
             return FT_Err_Invalid_Argument;
           }
@@ -462,7 +479,7 @@
             return FT_Err_Out_Of_Memory;
           }
 
-          if ( fread( font->file_address, 1, file_size, file ) != file_size )
+          if ( !fread( font->file_address, file_size, 1, file ) )
           {
             free( font->file_address );
             free( font );
@@ -480,32 +497,6 @@
           font->file_size    = 0;
         }
 
-        switch ( handle->encoding )
-        {
-        case FT_ENCODING_NONE:
-          font->num_indices = face->num_glyphs;
-          break;
-
-        case FT_ENCODING_UNICODE:
-          font->num_indices = 0x110000L;
-          break;
-
-        case FT_ENCODING_ADOBE_LATIN_1:
-        case FT_ENCODING_ADOBE_STANDARD:
-        case FT_ENCODING_ADOBE_EXPERT:
-        case FT_ENCODING_ADOBE_CUSTOM:
-        case FT_ENCODING_APPLE_ROMAN:
-          font->num_indices = 0x100L;
-          break;
-
-          /* some fonts use range 0x00-0x100, others have 0xF000-0xF0FF */
-        case FT_ENCODING_MS_SYMBOL:
-          font->num_indices = 0x10000L;
-
-        default:
-          font->num_indices = 0x10000L;
-        }
-
         FT_Done_Face( face );
         face = NULL;
 
@@ -539,10 +530,45 @@
   FTDemo_Set_Current_Font( FTDemo_Handle*  handle,
                            PFont           font )
   {
+    FT_Face  face;
+
+
     handle->current_font   = font;
     handle->scaler.face_id = (FTC_FaceID)font;
 
-    handle->string_reload = 1;
+    error = FTC_Manager_LookupFace( handle->cache_manager,
+                                    handle->scaler.face_id, &face );
+
+    if ( font->cmap_index < face->num_charmaps )
+      handle->encoding = face->charmaps[font->cmap_index]->encoding;
+    else
+      handle->encoding = FT_ENCODING_ORDER;
+
+    switch ( handle->encoding )
+    {
+    case FT_ENCODING_ORDER:
+      font->num_indices = face->num_glyphs;
+      break;
+
+    case FT_ENCODING_UNICODE:
+      font->num_indices = 0x110000L;
+      break;
+
+    case FT_ENCODING_ADOBE_LATIN_1:
+    case FT_ENCODING_ADOBE_STANDARD:
+    case FT_ENCODING_ADOBE_EXPERT:
+    case FT_ENCODING_ADOBE_CUSTOM:
+    case FT_ENCODING_APPLE_ROMAN:
+      font->num_indices = 0x100L;
+      break;
+
+    /* some fonts use range 0x00-0x100, others have 0xF000-0xF0FF */
+    case FT_ENCODING_MS_SYMBOL:
+      font->num_indices = 0x10000L;
+
+    default:
+      font->num_indices = 0x10000L;
+    }
   }
 
 
@@ -550,16 +576,40 @@
   FTDemo_Set_Current_Size( FTDemo_Handle*  handle,
                            int             pixel_size )
   {
+    FT_Face  face;
+
+
     if ( pixel_size > 0xFFFF )
       pixel_size = 0xFFFF;
 
+    error = FTC_Manager_LookupFace( handle->cache_manager,
+                                    handle->scaler.face_id, &face );
+
+    if ( !error && !FT_IS_SCALABLE ( face ) )
+    {
+      int  i, j = 0;
+      int  c, d = abs( (int)face->available_sizes[j].y_ppem - pixel_size * 64 );
+
+
+      for ( i = 1; i < face->num_fixed_sizes; i++ )
+      {
+        c = abs( (int)face->available_sizes[i].y_ppem - pixel_size * 64 );
+
+        if ( c < d )
+        {
+          d = c;
+          j = i;
+        }
+      }
+
+      pixel_size = face->available_sizes[j].y_ppem / 64 ;
+    }
+
     handle->scaler.width  = (FT_UInt)pixel_size;
     handle->scaler.height = (FT_UInt)pixel_size;
     handle->scaler.pixel  = 1;                  /* activate integer format */
     handle->scaler.x_res  = 0;
     handle->scaler.y_res  = 0;
-
-    handle->string_reload = 1;
   }
 
 
@@ -568,17 +618,42 @@
                                int             char_size,
                                int             resolution )
   {
+    FT_Face  face;
+
+
     /* in 26.6 format, corresponding to (almost) 0x4000ppem */
     if ( char_size > 0xFFFFF )
       char_size = 0xFFFFF;
 
+    error = FTC_Manager_LookupFace( handle->cache_manager,
+                                    handle->scaler.face_id, &face );
+
+    if ( !error && !FT_IS_SCALABLE ( face ) )
+    {
+      int  pixel_size = char_size * resolution / 72;
+      int  i, j = 0;
+      int  c, d = abs( (int)face->available_sizes[j].y_ppem - pixel_size );
+
+
+      for ( i = 1; i < face->num_fixed_sizes; i++ )
+      {
+        c = abs( (int)face->available_sizes[i].y_ppem - pixel_size );
+
+        if ( c < d )
+        {
+          d = c;
+          j = i;
+        }
+      }
+
+      char_size = face->available_sizes[j].y_ppem * 72 / resolution;
+    }
+
     handle->scaler.width  = (FT_UInt)char_size;
     handle->scaler.height = (FT_UInt)char_size;
     handle->scaler.pixel  = 0;                     /* activate 26.6 format */
     handle->scaler.x_res  = (FT_UInt)resolution;
     handle->scaler.y_res  = (FT_UInt)resolution;
-
-    handle->string_reload = 1;
   }
 
 
@@ -604,34 +679,37 @@
     if ( !handle->use_sbits )
       flags |= FT_LOAD_NO_BITMAP;
 
+    if ( handle->use_color )
+      flags |= FT_LOAD_COLOR;
+
     if ( handle->hinted )
     {
       target = 0;
 
-      if ( handle->antialias )
+      switch ( handle->lcd_mode )
       {
-        switch ( handle->lcd_mode )
-        {
-        case LCD_MODE_LIGHT:
-          target = FT_LOAD_TARGET_LIGHT;
-          break;
-
-        case LCD_MODE_RGB:
-        case LCD_MODE_BGR:
-          target = FT_LOAD_TARGET_LCD;
-          break;
-
-        case LCD_MODE_VRGB:
-        case LCD_MODE_VBGR:
-          target = FT_LOAD_TARGET_LCD_V;
-          break;
-
-        default:
-          target = FT_LOAD_TARGET_NORMAL;
-        }
-      }
-      else
+      case LCD_MODE_MONO:
         target = FT_LOAD_TARGET_MONO;
+        break;
+
+      case LCD_MODE_LIGHT:
+      case LCD_MODE_LIGHT_SUBPIXEL:
+        target = FT_LOAD_TARGET_LIGHT;
+        break;
+
+      case LCD_MODE_RGB:
+      case LCD_MODE_BGR:
+        target = FT_LOAD_TARGET_LCD;
+        break;
+
+      case LCD_MODE_VRGB:
+      case LCD_MODE_VBGR:
+        target = FT_LOAD_TARGET_LCD_V;
+        break;
+
+      default:
+        target = FT_LOAD_TARGET_NORMAL;
+      }
 
       flags |= target;
     }
@@ -639,15 +717,11 @@
     {
       flags |= FT_LOAD_NO_HINTING;
 
-      if ( !handle->antialias )
+      if ( handle->lcd_mode == LCD_MODE_MONO )
         flags |= FT_LOAD_MONOCHROME;
     }
 
-    if ( handle->color )
-      flags |= FT_LOAD_COLOR;
-
-    handle->load_flags    = flags;
-    handle->string_reload = 1;
+    handle->load_flags = flags;
   }
 
 
@@ -655,12 +729,17 @@
   FTDemo_Get_Index( FTDemo_Handle*  handle,
                     FT_UInt32       charcode )
   {
-    FTC_FaceID  face_id = handle->scaler.face_id;
-    PFont       font    = handle->current_font;
+    if ( handle->encoding != FT_ENCODING_ORDER )
+    {
+      FTC_FaceID  face_id = handle->scaler.face_id;
+      PFont       font    = handle->current_font;
 
 
-    return FTC_CMapCache_Lookup( handle->cmap_cache, face_id,
-                                 font->cmap_index, charcode );
+      return FTC_CMapCache_Lookup( handle->cmap_cache, face_id,
+                                   font->cmap_index, charcode );
+    }
+    else
+      return (FT_UInt)charcode;
   }
 
 
@@ -682,6 +761,170 @@
   }
 
 
+  void
+  FTDemo_Draw_Header( FTDemo_Handle*   handle,
+                      FTDemo_Display*  display,
+                      int              ptsize,
+                      int              res,
+                      int              idx,
+                      int              error_code )
+  {
+    FT_Face      face;
+    char         buf[256];
+    const char*  basename;
+    int          ppem;
+
+    int          line = 0;
+    int          x;
+
+
+    error = FTC_Manager_LookupFace( handle->cache_manager,
+                                    handle->scaler.face_id, &face );
+    if ( error )
+    {
+      FTDemo_Display_Done( display );
+      FTDemo_Done( handle );
+      PanicZ( "can't access font file" );
+    }
+
+
+    /* font and file name */
+    basename = ft_basename( handle->current_font->filepathname );
+    sprintf( buf, "%.50s %.50s (file `%.100s')",
+             face->family_name, face->style_name, basename );
+    grWriteCellString( display->bitmap, 0, (line++) * HEADER_HEIGHT,
+                       buf, display->fore_color );
+
+    /* ppem, pt and dpi, instance */
+    ppem = FT_IS_SCALABLE( face ) ? FT_MulFix( face->units_per_EM,
+                                               face->size->metrics.y_scale )
+                                  : face->size->metrics.y_ppem * 64;
+
+    if ( res == 72 )
+      x  = sprintf( buf, "%.4g ppem", ppem / 64.0 );
+    else
+      x  = sprintf( buf, "%g pt at %d dpi, %.4g ppem",
+                         ptsize / 64.0, res, ppem / 64.0 );
+
+    if ( face->face_index >> 16 )
+      x += sprintf( buf + x, ", instance %ld/%ld",
+                             face->face_index >> 16,
+                             face->style_flags >> 16 );
+
+    grWriteCellString( display->bitmap, 0, line * HEADER_HEIGHT,
+                       buf, display->fore_color );
+
+    if ( abs( ptsize * res / 64 - face->size->metrics.y_ppem * 72 ) > 36 ||
+         error_code                                                      )
+    {
+      switch ( error_code )
+      {
+      case FT_Err_Ok:
+        sprintf( buf, "Available size shown" );
+        break;
+      case FT_Err_Invalid_Pixel_Size:
+        sprintf( buf, "Invalid pixel size" );
+        break;
+      case FT_Err_Invalid_PPem:
+        sprintf( buf, "Invalid ppem value" );
+        break;
+      default:
+        sprintf( buf, "Error 0x%04x", (FT_UShort)error_code );
+      }
+      grWriteCellString( display->bitmap, 8 * x + 16, line * HEADER_HEIGHT,
+                         buf, display->warn_color );
+    }
+
+    /* gamma */
+    if ( display->gamma == 0.0 )
+      sprintf( buf, "gamma: sRGB" );
+    else
+      sprintf( buf, "gamma = %.1f", display->gamma );
+    grWriteCellString( display->bitmap,
+                       display->bitmap->width - 8 * 11, line * HEADER_HEIGHT,
+                       buf, display->fore_color );
+
+    line++;
+
+    /* encoding charcode or glyph index, glyph name */
+    if ( idx >= 0 )
+    {
+      const char*  encoding = NULL;
+      FT_UInt      glyph_idx = FTDemo_Get_Index( handle, (FT_UInt32)idx );
+
+
+      switch ( handle->encoding )
+      {
+      case FT_ENCODING_ORDER:
+        encoding = "glyph order";
+        break;
+      case FT_ENCODING_MS_SYMBOL:
+        encoding = "MS Symbol";
+        break;
+      case FT_ENCODING_UNICODE:
+        encoding = "Unicode";
+        break;
+      case FT_ENCODING_SJIS:
+        encoding = "SJIS";
+        break;
+      case FT_ENCODING_PRC:
+        encoding = "PRC";
+        break;
+      case FT_ENCODING_BIG5:
+        encoding = "Big5";
+        break;
+      case FT_ENCODING_WANSUNG:
+        encoding = "Wansung";
+        break;
+      case FT_ENCODING_JOHAB:
+        encoding = "Johab";
+        break;
+      case FT_ENCODING_ADOBE_STANDARD:
+        encoding = "Adobe Standard";
+        break;
+      case FT_ENCODING_ADOBE_EXPERT:
+        encoding = "Adobe Expert";
+        break;
+      case FT_ENCODING_ADOBE_CUSTOM:
+        encoding = "Adobe Custom";
+        break;
+      case FT_ENCODING_ADOBE_LATIN_1:
+        encoding = "Latin 1";
+        break;
+      case FT_ENCODING_OLD_LATIN_2:
+        encoding = "Latin 2";
+        break;
+      case FT_ENCODING_APPLE_ROMAN:
+        encoding = "Apple Roman";
+        break;
+      default:
+        encoding = "Other";
+      }
+
+      if ( handle->encoding == FT_ENCODING_ORDER )
+        x = sprintf( buf, "%s idx: %d",
+                          encoding, idx );
+      else if ( handle->encoding == FT_ENCODING_UNICODE )
+        x = sprintf( buf, "%s charcode: U+%04X (glyph idx %d)",
+                          encoding, idx, glyph_idx );
+      else
+        x = sprintf( buf, "%s charcode: 0x%X (glyph idx %d)",
+                          encoding, idx, glyph_idx );
+
+      if ( FT_HAS_GLYPH_NAMES( face ) )
+      {
+        x += sprintf( buf + x, ", name: " );
+
+        FT_Get_Glyph_Name( face, glyph_idx, buf + x, (FT_UInt)( 256 - x ) );
+      }
+
+      grWriteCellString( display->bitmap, 0, (line++) * HEADER_HEIGHT,
+                         buf, display->fore_color );
+    }
+
+  }
+
+
   FT_Error
   FTDemo_Glyph_To_Bitmap( FTDemo_Handle*  handle,
                           FT_Glyph        glyf,
@@ -702,19 +945,32 @@
 
     if ( glyf->format == FT_GLYPH_FORMAT_OUTLINE )
     {
-      FT_Render_Mode  render_mode = FT_RENDER_MODE_MONO;
+      FT_Render_Mode  render_mode;
 
 
-      if ( handle->antialias )
+      switch ( handle->lcd_mode )
       {
-        if ( handle->lcd_mode == 0 )
-          render_mode = FT_RENDER_MODE_NORMAL;
-        else if ( handle->lcd_mode == 1 )
-          render_mode = FT_RENDER_MODE_LIGHT;
-        else if ( handle->lcd_mode <= 3 )
-          render_mode = FT_RENDER_MODE_LCD;
-        else
-          render_mode = FT_RENDER_MODE_LCD_V;
+      case LCD_MODE_MONO:
+        render_mode = FT_RENDER_MODE_MONO;
+        break;
+
+      case LCD_MODE_LIGHT:
+      case LCD_MODE_LIGHT_SUBPIXEL:
+        render_mode = FT_RENDER_MODE_LIGHT;
+        break;
+
+      case LCD_MODE_RGB:
+      case LCD_MODE_BGR:
+        render_mode = FT_RENDER_MODE_LCD;
+        break;
+
+      case LCD_MODE_VRGB:
+      case LCD_MODE_VBGR:
+        render_mode = FT_RENDER_MODE_LCD_V;
+        break;
+
+      default:
+        render_mode = FT_RENDER_MODE_NORMAL;
       }
 
       /* render the glyph to a bitmap, don't destroy original */
@@ -758,14 +1014,16 @@
       break;
 
     case FT_PIXEL_MODE_LCD:
-      target->mode  = handle->lcd_mode == 2 ? gr_pixel_mode_lcd
-                                            : gr_pixel_mode_lcd2;
+      target->mode  = handle->lcd_mode == LCD_MODE_RGB
+                      ? gr_pixel_mode_lcd
+                      : gr_pixel_mode_lcd2;
       target->grays = source->num_grays;
       break;
 
     case FT_PIXEL_MODE_LCD_V:
-      target->mode  = handle->lcd_mode == 4 ? gr_pixel_mode_lcdv
-                                            : gr_pixel_mode_lcdv2;
+      target->mode  = handle->lcd_mode == LCD_MODE_VRGB
+                      ? gr_pixel_mode_lcdv
+                      : gr_pixel_mode_lcdv2;
       target->grays = source->num_grays;
       break;
 
@@ -867,13 +1125,15 @@
           break;
 
         case FT_PIXEL_MODE_LCD:
-          target->mode  = handle->lcd_mode == 2 ? gr_pixel_mode_lcd
-                                                : gr_pixel_mode_lcd2;
+          target->mode  = handle->lcd_mode == LCD_MODE_RGB
+                          ? gr_pixel_mode_lcd
+                          : gr_pixel_mode_lcd2;
           break;
 
         case FT_PIXEL_MODE_LCD_V:
-          target->mode  = handle->lcd_mode == 4 ? gr_pixel_mode_lcdv
-                                                : gr_pixel_mode_lcdv2;
+          target->mode  = handle->lcd_mode == LCD_MODE_VRGB
+                          ? gr_pixel_mode_lcdv
+                          : gr_pixel_mode_lcdv2;
           break;
 
         case FT_PIXEL_MODE_BGRA:
@@ -948,7 +1208,7 @@
     if ( glyf )
       FT_Done_Glyph( glyf );
 
-    *pen_x += x_advance + 1;
+    *pen_x += x_advance;
 
     return FT_Err_Ok;
   }
@@ -983,7 +1243,7 @@
     if ( glyf )
       FT_Done_Glyph( glyf );
 
-    *pen_x += x_advance + 1;
+    *pen_x += x_advance;
 
     return FT_Err_Ok;
   }
@@ -1047,10 +1307,7 @@
 
       codepoint = (unsigned long)ch;
 
-      if ( handle->encoding != FT_ENCODING_NONE )
-        glyph->glyph_index = FTDemo_Get_Index( handle, codepoint );
-      else
-        glyph->glyph_index = codepoint;
+      glyph->glyph_index = FTDemo_Get_Index( handle, codepoint );
 
       glyph++;
       handle->string_length++;
@@ -1058,18 +1315,19 @@
       if ( handle->string_length >= MAX_GLYPHS )
         break;
     }
-
-    handle->string_reload = 1;
   }
 
 
-  static FT_Error
-  string_load( FTDemo_Handle*  handle )
+  FT_Error
+  FTDemo_String_Load( FTDemo_Handle*          handle,
+                      FTDemo_String_Context*  sc )
   {
-    int      n;
     FT_Size  size;
     FT_Face  face;
-    FT_Pos   prev_rsb_delta = 0;
+    FT_Int   i;
+    FT_Int   length = handle->string_length;
+    PGlyph   glyph, prev;
+    FT_Pos   track_kern   = 0;
 
 
     error = FTDemo_Get_Size( handle, &size );
@@ -1078,11 +1336,8 @@
 
     face = size->face;
 
-    for ( n = 0; n < handle->string_length; n++ )
+    for ( glyph = handle->string, i = 0; i < length; glyph++, i++ )
     {
-      PGlyph  glyph = handle->string + n;
-
-
       /* clear existing image if there is one */
       if ( glyph->image )
       {
@@ -1106,192 +1361,133 @@
         glyph->vadvance.x = 0;
         glyph->vadvance.y = -metrics->vertAdvance;
 
-        if ( prev_rsb_delta - face->glyph->lsb_delta >= 32 )
-          glyph->delta = -1 * 64;
-        else if ( prev_rsb_delta - face->glyph->lsb_delta < -32 )
-          glyph->delta = 1 * 64;
-        else
-          glyph->delta = 0;
+        glyph->lsb_delta = face->glyph->lsb_delta;
+        glyph->rsb_delta = face->glyph->rsb_delta;
+
+        glyph->hadvance.x = metrics->horiAdvance;
+        glyph->hadvance.y = 0;
       }
     }
 
-    return FT_Err_Ok;
-  }
-
-
-  static FT_Error
-  string_render_prepare( FTDemo_Handle*          handle,
-                         FTDemo_String_Context*  sc,
-                         FT_Vector*              advances )
-  {
-    FT_Face     face;
-    FT_Size     size;
-    PGlyph      glyph;
-    FT_Pos      track_kern   = 0;
-    FT_UInt     prev_index   = 0;
-    FT_Vector*  prev_advance = NULL;
-    FT_Vector   extent       = { 0, 0 };
-    FT_Int      i;
-
-
-    error = FTDemo_Get_Size( handle, &size );
-    if ( error )
-      return error;
-
-    face = size->face;
-
-    if ( !sc->vertical && sc->kerning_degree )
+    if ( sc->kerning_degree )
     {
       /* this function needs and returns points, not pixels */
-      if ( FT_Get_Track_Kerning( face,
-                                 (FT_Fixed)handle->scaler.width << 10,
-                                 -sc->kerning_degree,
-                                 &track_kern ) )
-        track_kern = 0;
-      else
+      if ( !FT_Get_Track_Kerning( face,
+                                  (FT_Fixed)handle->scaler.width << 10,
+                                  -sc->kerning_degree,
+                                  &track_kern ) )
         track_kern = (FT_Pos)(
                        ( track_kern / 1024.0 * handle->scaler.x_res ) /
                        72.0 );
     }
 
-    for ( i = 0; i < handle->string_length; i++ )
+    for ( prev = handle->string + length, glyph = handle->string, i = 0;
+          i < length;
+          prev = glyph, glyph++, i++ )
     {
-      glyph = handle->string + i;
-
       if ( !glyph->image )
         continue;
 
-      if ( sc->vertical )
-        advances[i] = glyph->vadvance;
-      else
+      if ( handle->lcd_mode == LCD_MODE_LIGHT_SUBPIXEL )
+        glyph->hadvance.x += glyph->lsb_delta - glyph->rsb_delta;
+
+      prev->hadvance.x += track_kern;
+
+      if ( sc->kerning_mode )
       {
-        advances[i]     = glyph->image->advance;
-        advances[i].x >>= 10;
-        advances[i].y >>= 10;
+        FT_Vector  kern;
 
-        if ( prev_advance )
+
+        FT_Get_Kerning( face, prev->glyph_index, glyph->glyph_index,
+                        FT_KERNING_UNFITTED, &kern );
+
+        prev->hadvance.x += kern.x;
+        prev->hadvance.y += kern.y;
+
+        if ( handle->lcd_mode != LCD_MODE_LIGHT_SUBPIXEL &&
+             sc->kerning_mode > KERNING_MODE_NORMAL      )
         {
-          prev_advance->x += track_kern;
-
-          if ( sc->kerning_mode )
-          {
-            FT_Vector  kern;
-
-
-            FT_Get_Kerning( face, prev_index, glyph->glyph_index,
-                FT_KERNING_UNFITTED, &kern );
-
-            prev_advance->x += kern.x;
-            prev_advance->y += kern.y;
-
-            if ( sc->kerning_mode > KERNING_MODE_NORMAL )
-              prev_advance->x += glyph->delta;
-          }
+          if ( prev->rsb_delta - glyph->lsb_delta > 32 )
+            prev->hadvance.x -= 64;
+          else if ( prev->rsb_delta - glyph->lsb_delta < -31 )
+            prev->hadvance.x += 64;
         }
       }
 
-      if ( prev_advance )
+      if ( handle->lcd_mode != LCD_MODE_LIGHT_SUBPIXEL &&
+           handle->hinted                              )
       {
-        if ( handle->hinted )
-        {
-          prev_advance->x = ROUND( prev_advance->x );
-          prev_advance->y = ROUND( prev_advance->y );
-        }
-
-        extent.x += prev_advance->x;
-        extent.y += prev_advance->y;
+        prev->hadvance.x = ROUND( prev->hadvance.x );
+        prev->hadvance.y = ROUND( prev->hadvance.y );
       }
-
-      prev_index   = glyph->glyph_index;
-      prev_advance = advances + i;
     }
 
-    if ( prev_advance )
-    {
-      if ( handle->hinted )
-      {
-        prev_advance->x = ROUND( prev_advance->x );
-        prev_advance->y = ROUND( prev_advance->y );
-      }
-
-      extent.x += prev_advance->x;
-      extent.y += prev_advance->y;
-    }
-
-    /* store the extent in the last slot */
-    i = handle->string_length - 1;
-    advances[i] = extent;
-
     return FT_Err_Ok;
   }
 
 
-  FT_Error
+  int
   FTDemo_String_Draw( FTDemo_Handle*          handle,
                       FTDemo_Display*         display,
                       FTDemo_String_Context*  sc,
                       int                     x,
                       int                     y )
   {
-    int        n;
-    FT_Vector  pen, advances[MAX_GLYPHS];
-    FT_Size    size;
-    FT_Face    face;
+    int        first = sc->offset;
+    int        last  = handle->string_length;
+    int        m, n;
+    FT_Vector  pen = { 0, 0};
+    FT_Vector  advance;
 
 
-    if ( !sc                        ||
-         x < 0                      ||
+    if ( x < 0                      ||
          y < 0                      ||
          x > display->bitmap->width ||
          y > display->bitmap->rows  )
-      return FT_Err_Invalid_Argument;
-
-    error = FTDemo_Get_Size( handle, &size );
-    if ( error )
-      return error;
-
-    face = size->face;
-
-    if ( handle->string_reload )
-    {
-      error = string_load( handle );
-      if ( error )
-        return error;
-
-      handle->string_reload = 0;
-    }
-
-    error = string_render_prepare( handle, sc, advances );
-    if ( error )
-      return error;
+      return 0;
 
     /* change to Cartesian coordinates */
     y = display->bitmap->rows - y;
 
-    /* get the extent, which we store in the last slot */
-    pen = advances[handle->string_length - 1];
+    /* calculate the extent */
+    if ( sc->extent )
+      for( n = first; n < first + last || pen.x > 0; n++ )  /* chk progress */
+      {
+        m = n % handle->string_length;  /* recycling */
+        if ( pen.x + handle->string[m].hadvance.x > sc->extent )
+        {
+          last = n;
+          break;
+        }
+        pen.x += handle->string[m].hadvance.x;
+        pen.y += handle->string[m].hadvance.y;
+      }
+    else if ( sc->vertical )
+      for ( n = first; n < last; n++ )
+      {
+        pen.x += handle->string[n].vadvance.x;
+        pen.y += handle->string[n].vadvance.y;
+      }
+    else
+      for ( n = first; n < last; n++ )
+      {
+        pen.x += handle->string[n].hadvance.x;
+        pen.y += handle->string[n].hadvance.y;
+      }
 
     pen.x = FT_MulFix( pen.x, sc->center );
     pen.y = FT_MulFix( pen.y, sc->center );
 
     /* XXX sbits */
     /* get pen position */
-    if ( sc->matrix && FT_IS_SCALABLE( face ) )
-    {
-      FT_Vector_Transform( &pen, sc->matrix );
+    FT_Vector_Transform( &pen, sc->matrix );
 
-      pen.x = ( x << 6 ) - pen.x;
-      pen.y = ( y << 6 ) - pen.y;
-    }
-    else
-    {
-      pen.x = ROUND( ( x << 6 ) - pen.x );
-      pen.y = ROUND( ( y << 6 ) - pen.y );
-    }
+    pen.x = ( x << 6 ) - pen.x;
+    pen.y = ( y << 6 ) - pen.y;
 
-    for ( n = 0; n < handle->string_length; n++ )
+    for ( n = first; n < last; n++ )
     {
-      PGlyph    glyph = handle->string + n;
+      PGlyph    glyph = handle->string + n % handle->string_length;
       FT_Glyph  image;
       FT_BBox   bbox;
 
@@ -1335,11 +1531,13 @@
         }
       }
 
-      if ( sc->matrix )
-        FT_Vector_Transform( advances + n, sc->matrix );
+      advance = sc->vertical ? glyph->vadvance : glyph->hadvance;
 
-      pen.x += advances[n].x;
-      pen.y += advances[n].y;
+      if ( sc->matrix )
+        FT_Vector_Transform( &advance, sc->matrix );
+
+      pen.x += advance.x;
+      pen.y += advance.y;
 
       FT_Glyph_Get_CBox( image, FT_GLYPH_BBOX_PIXELS, &bbox );
 
@@ -1382,11 +1580,11 @@
       FT_Done_Glyph( image );
     }
 
-    return error;
+    return last - first;
   }
 
 
-  FT_Encoding
+  unsigned long
   FTDemo_Make_Encoding_Tag( const char*  s )
   {
     int            i;
@@ -1401,7 +1599,91 @@
       l  += (unsigned long)s[i];
     }
 
-    return (FT_Encoding)l;
+    return l;
+  }
+
+
+  int
+  FTDemo_Event_Cff_Hinting_Engine_Change( FT_Library     library,
+                                          unsigned int*  current,
+                                          unsigned int   delta )
+  {
+    unsigned int  new_cff_hinting_engine = 0;
+
+
+    if ( delta )
+      new_cff_hinting_engine = ( *current          +
+                                 delta             +
+                                 N_HINTING_ENGINES ) % N_HINTING_ENGINES;
+
+    error = FT_Property_Set( library,
+                             "cff",
+                             "hinting-engine",
+                             &new_cff_hinting_engine );
+
+    if ( !error )
+    {
+      *current = new_cff_hinting_engine;
+      return 1;
+    }
+
+    return 0;
+  }
+
+
+  int
+  FTDemo_Event_Type1_Hinting_Engine_Change( FT_Library     library,
+                                            unsigned int*  current,
+                                            unsigned int   delta )
+  {
+    unsigned int  new_type1_hinting_engine = 0;
+
+
+    if ( delta )
+      new_type1_hinting_engine = ( *current          +
+                                   delta             +
+                                   N_HINTING_ENGINES ) % N_HINTING_ENGINES;
+
+    error = FT_Property_Set( library,
+                             "type1",
+                             "hinting-engine",
+                             &new_type1_hinting_engine );
+
+    if ( !error )
+    {
+      *current = new_type1_hinting_engine;
+      return 1;
+    }
+
+    return 0;
+  }
+
+
+  int
+  FTDemo_Event_T1cid_Hinting_Engine_Change( FT_Library     library,
+                                            unsigned int*  current,
+                                            unsigned int   delta )
+  {
+    unsigned int  new_t1cid_hinting_engine = 0;
+
+
+    if ( delta )
+      new_t1cid_hinting_engine = ( *current          +
+                                   delta             +
+                                   N_HINTING_ENGINES ) % N_HINTING_ENGINES;
+
+    error = FT_Property_Set( library,
+                             "t1cid",
+                             "hinting-engine",
+                             &new_t1cid_hinting_engine );
+
+    if ( !error )
+    {
+      *current = new_t1cid_hinting_engine;
+      return 1;
+    }
+
+    return 0;
   }
 
 
